#
# Clausification module
# clausification (polynomial-spaced with introducing new predicates for rule bodies) with/without
# Skolem functions
#
# Orkunt Sabuncu
#

from completion import CompletedPart, CompletedFormula
from atomrule import *
from copy import deepcopy, copy

'''
    Does the formula needs aux predicates for each disjunctive part? 
    Is one disjunction in the clausified formula enough ?
'''
def doesFormulaNeedsAuxPredicates(compformula):
    if not compformula.predicate:
	return (False, False)
    if len(compformula.disj_part) < 2:
	# we assume formula has at least one disj_part !!
	if len(compformula.disj_part[0].conj_part) > 1:
	    return (False, False)
	else:
	    return (False, True)
    for cp in compformula.disj_part:
	if len(cp.conj_part) > 1:
	    return (True,False)
    return (False,True)
    
def clausetoTPTPFormat(clause, name='cl'):
    str = 'cnf('+name+',axiom,(\n  '
    tmp_str = ''
    for l in clause:
	tmp_str += '\t'+ l.toTPTPFormat() + ' |\n'
    tmp_str = tmp_str[:-3] + ' )).'
    str += tmp_str
    return str
    
def change_to_skolems(l, var_to_sko):
    for i,v in enumerate(l.vars):
	if v[0] == 'V' and v[1] in var_to_sko:
	    l.vars[i] = var_to_sko[v[1]]
    
def clausify_formula_definitional_withsko(compformula):
    needs_auxpred, onedisj_enough = doesFormulaNeedsAuxPredicates(compformula)
    clauses = []
    
    # a constraint
    if (not compformula.predicate):
        tmp_claus = []
        tmp_claus.extend(deepcopy(compformula.disj_part[0].conj_part))
        for l in tmp_claus:
            l.naf = not l.naf
        clauses.append(tmp_claus)
        return clauses
    
    # <= part
    if not needs_auxpred:
        for cp in compformula.disj_part:
            tmp_claus = [compformula.predicate]
            tmp_claus.extend(deepcopy(cp.conj_part))
            for l in tmp_claus[1:]:
                l.naf = not l.naf
            clauses.append(tmp_claus)
    
    # => part (but, if it needs_auxpred <= part is also generated)
    sko_idx = 1
    sko_uvars = ''
    for v in compformula.predicate.vars:
        sko_uvars += v[1]+','
    sko_uvars = sko_uvars[:-1]
    
    if onedisj_enough:
        # all ... <- comp_pred
        tmp_claus = [copy(compformula.predicate)]
        tmp_claus[0].naf = True
        for cp in compformula.disj_part:
            existentials_to_skolems = {}
            if len(cp.existential_vars) > 0:
                for v in cp.existential_vars:
                    tmp_str = ''
                    if sko_uvars != '':
                        #tmp_str = '('+sko_uvars+','+v[1]+')'
                        tmp_str = '('+sko_uvars+')'
                    existentials_to_skolems[v[1]] = ('O','sko_%s%d%s' %(compformula.predicate.atom.name, sko_idx, tmp_str))
                    sko_idx += 1
                    
            tmp_claus.append(cp.conj_part[0])
            if len(cp.existential_vars) > 0:
                change_to_skolems(tmp_claus[-1],existentials_to_skolems)
        clauses.append(tmp_claus)
    else:
        if needs_auxpred:
            # aux1 v aux2 v ... <- comp_pred
            tmp_claus = [copy(compformula.predicate)]
            tmp_claus[0].naf = True
            for i,cp in enumerate(compformula.disj_part):
                tmp_atom = Atom('%s_r%d' % (compformula.predicate.atom.name, i+1),compformula.predicate.atom.arity)
                tmp_lit = Literal(tmp_atom, False, compformula.predicate.vars)
                tmp_claus.append(tmp_lit)
                
                # p(..) <- aux1
                tmp_claus_definitional = [copy(compformula.predicate)]
                tmp_lit_definitional = copy(tmp_lit)
                tmp_lit_definitional.naf = True
                tmp_claus_definitional.append(tmp_lit_definitional)
                clauses.append(tmp_claus_definitional)
                
                # aux1 <- b(..) & ~b1(..) & ...
                tmp_claus_definitional_2 = [copy(tmp_lit)]
                tmp_claus_definitional_2.extend(deepcopy(cp.conj_part))
                for l in tmp_claus_definitional_2[1:]:
                    l.naf = not l.naf
                clauses.append(tmp_claus_definitional_2)
                
                # ... <- aux1 
                existentials_to_skolems = {}
                if len(cp.existential_vars) > 0:
                    for v in cp.existential_vars:
                        tmp_str = ''
                        if sko_uvars != '':
                            #tmp_str = '('+sko_uvars+','+v[1]+')'
                            tmp_str = '('+sko_uvars+')'
                        existentials_to_skolems[v[1]] = ('O','sko_%s%d%s' %(compformula.predicate.atom.name, sko_idx, tmp_str))
                        sko_idx += 1
                        
                tmp_lit2 = copy(tmp_lit)
                tmp_lit2.naf = True
                
                for l in cp.conj_part:
                    tmp_claus2 = [tmp_lit2]
                    tmp_claus2.append(deepcopy(l))
                    if len(cp.existential_vars) > 0:
                        change_to_skolems(tmp_claus2[-1],existentials_to_skolems)
                    clauses.append(tmp_claus2)
                    
            clauses.append(tmp_claus)
        else:
            # ... <- comp_pred
            for cp in compformula.disj_part:
                existentials_to_skolems = {}
                if len(cp.existential_vars) > 0:
                    for v in cp.existential_vars:
                        tmp_str = ''
                        if sko_uvars != '':
                            #tmp_str = '('+sko_uvars+','+v[1]+')'
                            tmp_str = '('+sko_uvars+')'
                        existentials_to_skolems[v[1]] = ('O','sko_%s%d%s' %(compformula.predicate.atom.name, sko_idx, tmp_str))
                        sko_idx += 1
            
                tmp_lit = copy(compformula.predicate)
                tmp_lit.naf = True
                for l in cp.conj_part:
                    tmp_claus = [tmp_lit]
                    tmp_claus.append(deepcopy(l))
                    if len(cp.existential_vars) > 0:
                        change_to_skolems(tmp_claus[-1],existentials_to_skolems)
                    clauses.append(tmp_claus)
    
    # removed for definitional clausification approach
    # r(..) <- b(..) & ~b1(..) & ...
    '''
    for cp in compformula.disj_part:
        tmp_claus = [compformula.predicate]
        tmp_claus.extend(deepcopy(cp.conj_part))
        for l in tmp_claus[1:]:
            l.naf = not l.naf
        clauses.append(tmp_claus)
    '''
                
    return clauses
    
def clausify_formula_definitional_withoutsko(compformula, sorts_inf):
    needs_auxpred, onedisj_enough = doesFormulaNeedsAuxPredicates(compformula)
    clauses = []
    
    # a constraint
    if (not compformula.predicate):
        tmp_claus = []
        tmp_claus.extend(deepcopy(compformula.disj_part[0].conj_part))
        for l in tmp_claus:
            l.naf = not l.naf
        clauses.append(tmp_claus)
        return clauses
    
    # <= part
    if not needs_auxpred:
        for cp in compformula.disj_part:
            tmp_claus = [compformula.predicate]
            tmp_claus.extend(deepcopy(cp.conj_part))
            for l in tmp_claus[1:]:
                l.naf = not l.naf
            clauses.append(tmp_claus)
        
    # => part (but, if it needs_auxpred <= part is also generated)
    sko_idx = 1
    sko_uvars = compformula.predicate.vars
    exs_sorts = {}
    
    if onedisj_enough:
        # all ... <- comp_pred
        tmp_claus = [copy(compformula.predicate)]
        tmp_claus[0].naf = True
        for cp in compformula.disj_part:
            existentials_to_skolempreds = {}
            if len(cp.existential_vars) > 0:
                for v in cp.existential_vars:
                    tmp_vars = sko_uvars + [v]
                    tmp_atom = Atom('skopred_%s%d' % (compformula.predicate.atom.name, sko_idx), len(tmp_vars))
                    tmp_lit = Literal(tmp_atom, True, tmp_vars)
                    existentials_to_skolempreds[v[1]] = tmp_lit
                    sko_idx += 1
                    
            tmp_claus.append(cp.conj_part[0])
            if len(cp.existential_vars) > 0:
                lst = [(existentials_to_skolempreds[v[1]],(cp.conj_part[0].atom.name,i+1)) for i,v in enumerate(cp.conj_part[0].vars) if v in cp.existential_vars]
                for (p,s) in lst:
                    if p not in exs_sorts:
                        exs_sorts[p] = s
                tmp_claus.extend(existentials_to_skolempreds.values())
        clauses.append(tmp_claus)
    else:
        if needs_auxpred:
            # aux1 v aux2 v ... <- comp_pred
            tmp_claus = [copy(compformula.predicate)]
            tmp_claus[0].naf = True
            for i,cp in enumerate(compformula.disj_part):
                tmp_atom = Atom('%s_r%d' % (compformula.predicate.atom.name, i+1),compformula.predicate.atom.arity)
                tmp_auxlit = Literal(tmp_atom, False, compformula.predicate.vars)
                tmp_claus.append(tmp_auxlit)
                
                # p(...) <- aux1
                tmp_claus_definitional = [copy(compformula.predicate)]
                tmp_lit_definitional = copy(tmp_auxlit)
                tmp_lit_definitional.naf = True
                tmp_claus_definitional.append(tmp_lit_definitional)
                clauses.append(tmp_claus_definitional)
                
                # aux1 <- b(..) & ~b1(..) & ...
                tmp_claus_definitional_2 = [copy(tmp_auxlit)]
                tmp_claus_definitional_2.extend(deepcopy(cp.conj_part))
                for l in tmp_claus_definitional_2[1:]:
                    l.naf = not l.naf
                clauses.append(tmp_claus_definitional_2)
                
                # ... <- aux1 
                existentials_to_skolempreds = {}
                if len(cp.existential_vars) > 0:
                    for v in cp.existential_vars:
                        tmp_vars = sko_uvars + [v]
                        tmp_atom = Atom('skopred_%s%d' % (compformula.predicate.atom.name, sko_idx), len(tmp_vars))
                        tmp_lit = Literal(tmp_atom, True, tmp_vars)
                        existentials_to_skolempreds[v[1]] = tmp_lit
                        sko_idx += 1
                        
                tmp_lit2 = copy(tmp_auxlit)
                tmp_lit2.naf = True
                
                for l in cp.conj_part:
                    tmp_claus2 = [tmp_lit2]
                    tmp_claus2.append(deepcopy(l))
                    tmp_claus2.extend([existentials_to_skolempreds[v[1]] for v in l.vars if v in cp.existential_vars])
                    lst = [(existentials_to_skolempreds[v[1]],(l.atom.name,i+1)) for i,v in enumerate(l.vars) if v in cp.existential_vars]
                    for (p,s) in lst:
                        if p not in exs_sorts:
                            exs_sorts[p] = s
                    clauses.append(tmp_claus2)
                    
            clauses.append(tmp_claus)
        else:
            # ... <- comp_pred
            for cp in compformula.disj_part:
                existentials_to_skolempreds = {}
                if len(cp.existential_vars) > 0:
                    for v in cp.existential_vars:
                        tmp_vars = sko_uvars + [v]
                        tmp_atom = Atom('skopred_%s%d' % (compformula.predicate.atom.name, sko_idx), len(tmp_vars))
                        tmp_lit = Literal(tmp_atom, True, tmp_vars)
                        existentials_to_skolempreds[v[1]] = tmp_lit
                        sko_idx += 1
            
                tmp_lit = copy(compformula.predicate)
                tmp_lit.naf = True
                for l in cp.conj_part:
                    tmp_claus = [tmp_lit]
                    tmp_claus.append(deepcopy(l))
                    tmp_claus.extend([existentials_to_skolempreds[v[1]] for v in l.vars if v in cp.existential_vars])
                    lst = [(existentials_to_skolempreds[v[1]],(l.atom.name,i+1)) for i,v in enumerate(l.vars) if v in cp.existential_vars]
                    for (p,s) in lst:
                        if p not in exs_sorts:
                            exs_sorts[p] = s
                    clauses.append(tmp_claus)
                
    if len(exs_sorts) > 0:
        clauses.extend(construct_totality_and_uniqueness_axioms(exs_sorts,sorts_inf))
    return clauses
    
    
def clausify_formula_withsko(compformula):
    needs_auxpred, onedisj_enough = doesFormulaNeedsAuxPredicates(compformula)
    clauses = []
    
    # a constraint
    if (not compformula.predicate):
	tmp_claus = []
	tmp_claus.extend(deepcopy(compformula.disj_part[0].conj_part))
	for l in tmp_claus:
	    l.naf = not l.naf
	clauses.append(tmp_claus)
	return clauses
    
    # <= part
    for cp in compformula.disj_part:
	tmp_claus = [compformula.predicate]
	tmp_claus.extend(deepcopy(cp.conj_part))
	for l in tmp_claus[1:]:
	    l.naf = not l.naf
	clauses.append(tmp_claus)
	
    # => part
    sko_idx = 1
    sko_uvars = ''
    for v in compformula.predicate.vars:
	sko_uvars += v[1]+','
    sko_uvars = sko_uvars[:-1]
    
    if onedisj_enough:
	# all ... <- comp_pred
        tmp_claus = [copy(compformula.predicate)]
	tmp_claus[0].naf = True
	for cp in compformula.disj_part:
            existentials_to_skolems = {}
	    if len(cp.existential_vars) > 0:
		for v in cp.existential_vars:
		    tmp_str = ''
		    if sko_uvars != '':
			#tmp_str = '('+sko_uvars+','+v[1]+')'
			tmp_str = '('+sko_uvars+')'
		    existentials_to_skolems[v[1]] = ('O','sko_%s%d%s' %(compformula.predicate.atom.name, sko_idx, tmp_str))
		    sko_idx += 1
		    
            tmp_claus.append(cp.conj_part[0])
	    if len(cp.existential_vars) > 0:
                change_to_skolems(tmp_claus[-1],existentials_to_skolems)
	clauses.append(tmp_claus)
    else:
	if needs_auxpred:
            # aux1 v aux2 v ... <- comp_pred
	    tmp_claus = [copy(compformula.predicate)]
	    tmp_claus[0].naf = True
	    for i,cp in enumerate(compformula.disj_part):
		tmp_atom = Atom('%s_r%d' % (compformula.predicate.atom.name, i+1),compformula.predicate.atom.arity)
		tmp_lit = Literal(tmp_atom, False, compformula.predicate.vars)
		tmp_claus.append(tmp_lit)
		
		# ... <- aux1 
		existentials_to_skolems = {}
                if len(cp.existential_vars) > 0:
		    for v in cp.existential_vars:
			tmp_str = ''
			if sko_uvars != '':
			    #tmp_str = '('+sko_uvars+','+v[1]+')'
			    tmp_str = '('+sko_uvars+')'
			existentials_to_skolems[v[1]] = ('O','sko_%s%d%s' %(compformula.predicate.atom.name, sko_idx, tmp_str))
			sko_idx += 1
			
		tmp_lit2 = copy(tmp_lit)
		tmp_lit2.naf = True
		
		for l in cp.conj_part:
		    tmp_claus2 = [tmp_lit2]
		    tmp_claus2.append(deepcopy(l))
		    if len(cp.existential_vars) > 0:
                        change_to_skolems(tmp_claus2[-1],existentials_to_skolems)
                    clauses.append(tmp_claus2)
		    
	    clauses.append(tmp_claus)
	else:
	    # ... <- comp_pred
	    for cp in compformula.disj_part:
                existentials_to_skolems = {}
		if len(cp.existential_vars) > 0:
		    for v in cp.existential_vars:
			tmp_str = ''
			if sko_uvars != '':
			    #tmp_str = '('+sko_uvars+','+v[1]+')'
			    tmp_str = '('+sko_uvars+')'
			existentials_to_skolems[v[1]] = ('O','sko_%s%d%s' %(compformula.predicate.atom.name, sko_idx, tmp_str))
			sko_idx += 1
	    
		tmp_lit = copy(compformula.predicate)
		tmp_lit.naf = True
		for l in cp.conj_part:
		    tmp_claus = [tmp_lit]
		    tmp_claus.append(deepcopy(l))
		    if len(cp.existential_vars) > 0:
			change_to_skolems(tmp_claus[-1],existentials_to_skolems)
                    clauses.append(tmp_claus)
		
    return clauses
    
def construct_totality_and_uniqueness_axioms(exs_sorts, sorts_inf):
    axioms = []

    #print exs_sorts
    #print sorts_inf
    
    for skolempred in exs_sorts.keys():
	sort = exs_sorts[skolempred]
	tot_claus = []
	for o in sorts_inf[0][sorts_inf[1][sort]]:
	    tmp_lit = Literal(skolempred.atom, False, skolempred.vars[:-1]+[('O',o)])
	    tot_claus.append(tmp_lit)
	axioms.append(tot_claus)
	
	eq_a = Atom('==',2)
	uniq_claus = [skolempred]
	lastvar = skolempred.vars[-1]
	lastvar2 = ('V',lastvar[1]+'1')
	skolempred2 = Literal(skolempred.atom, True, skolempred.vars[:-1] + [lastvar2])
	uniq_claus.append(skolempred2)
	eqlit = Literal(eq_a, False, [lastvar,lastvar2])
	uniq_claus.append(eqlit)
	axioms.append(uniq_claus)
	
    return axioms
	

def clausify_formula_withoutsko(compformula, sorts_inf):
    needs_auxpred, onedisj_enough = doesFormulaNeedsAuxPredicates(compformula)
    clauses = []
    
    # a constraint
    if (not compformula.predicate):
	tmp_claus = []
	tmp_claus.extend(deepcopy(compformula.disj_part[0].conj_part))
	for l in tmp_claus:
	    l.naf = not l.naf
	clauses.append(tmp_claus)
	return clauses
    
    # <= part
    for cp in compformula.disj_part:
	tmp_claus = [compformula.predicate]
	tmp_claus.extend(deepcopy(cp.conj_part))
	for l in tmp_claus[1:]:
	    l.naf = not l.naf
	clauses.append(tmp_claus)
	
    # => part
    sko_idx = 1
    sko_uvars = compformula.predicate.vars
    exs_sorts = {}
    
    if onedisj_enough:
	# all ... <- comp_pred
        tmp_claus = [copy(compformula.predicate)]
	tmp_claus[0].naf = True
	for cp in compformula.disj_part:
            existentials_to_skolempreds = {}
	    if len(cp.existential_vars) > 0:
		for v in cp.existential_vars:
		    tmp_vars = sko_uvars + [v]
		    tmp_atom = Atom('skopred_%s%d' % (compformula.predicate.atom.name, sko_idx), len(tmp_vars))
		    tmp_lit = Literal(tmp_atom, True, tmp_vars)
		    existentials_to_skolempreds[v[1]] = tmp_lit
		    sko_idx += 1
		    
            tmp_claus.append(cp.conj_part[0])
            if len(cp.existential_vars) > 0:
		lst = [(existentials_to_skolempreds[v[1]],(cp.conj_part[0].atom.name,i+1)) for i,v in enumerate(cp.conj_part[0].vars) if v in cp.existential_vars]
		for (p,s) in lst:
		    if p not in exs_sorts:
			exs_sorts[p] = s
                tmp_claus.extend(existentials_to_skolempreds.values())
	clauses.append(tmp_claus)
    else:
	if needs_auxpred:
            # aux1 v aux2 v ... <- comp_pred
	    tmp_claus = [copy(compformula.predicate)]
	    tmp_claus[0].naf = True
	    for i,cp in enumerate(compformula.disj_part):
		tmp_atom = Atom('%s_r%d' % (compformula.predicate.atom.name, i+1),compformula.predicate.atom.arity)
		tmp_auxlit = Literal(tmp_atom, False, compformula.predicate.vars)
		tmp_claus.append(tmp_auxlit)
		
		# ... <- aux1 
		existentials_to_skolempreds = {}
                if len(cp.existential_vars) > 0:
		    for v in cp.existential_vars:
			tmp_vars = sko_uvars + [v]
			tmp_atom = Atom('skopred_%s%d' % (compformula.predicate.atom.name, sko_idx), len(tmp_vars))
			tmp_lit = Literal(tmp_atom, True, tmp_vars)
			existentials_to_skolempreds[v[1]] = tmp_lit
			sko_idx += 1
			
		tmp_lit2 = copy(tmp_auxlit)
		tmp_lit2.naf = True
		
		for l in cp.conj_part:
		    tmp_claus2 = [tmp_lit2]
		    tmp_claus2.append(deepcopy(l))
		    tmp_claus2.extend([existentials_to_skolempreds[v[1]] for v in l.vars if v in cp.existential_vars])
		    lst = [(existentials_to_skolempreds[v[1]],(l.atom.name,i+1)) for i,v in enumerate(l.vars) if v in cp.existential_vars]
		    for (p,s) in lst:
                        if p not in exs_sorts:
                            exs_sorts[p] = s
		    clauses.append(tmp_claus2)
		    
	    clauses.append(tmp_claus)
	else:
	    # ... <- comp_pred
	    for cp in compformula.disj_part:
                existentials_to_skolempreds = {}
		if len(cp.existential_vars) > 0:
		    for v in cp.existential_vars:
			tmp_vars = sko_uvars + [v]
			tmp_atom = Atom('skopred_%s%d' % (compformula.predicate.atom.name, sko_idx), len(tmp_vars))
			tmp_lit = Literal(tmp_atom, True, tmp_vars)
			existentials_to_skolempreds[v[1]] = tmp_lit
			sko_idx += 1
	    
		tmp_lit = copy(compformula.predicate)
		tmp_lit.naf = True
		for l in cp.conj_part:
		    tmp_claus = [tmp_lit]
		    tmp_claus.append(deepcopy(l))
		    tmp_claus.extend([existentials_to_skolempreds[v[1]] for v in l.vars if v in cp.existential_vars])
		    lst = [(existentials_to_skolempreds[v[1]],(l.atom.name,i+1)) for i,v in enumerate(l.vars) if v in cp.existential_vars]
		    for (p,s) in lst:
			if p not in exs_sorts:
			    exs_sorts[p] = s
		    clauses.append(tmp_claus)
		
    if len(exs_sorts) > 0:
	clauses.extend(construct_totality_and_uniqueness_axioms(exs_sorts,sorts_inf))
    return clauses
    
    
