#
# Clark's completion module
# A conjunctive part for each rule is formed. A completion of a predicate is disjunctions
# of those conjunctive parts of the generator rules.
#
# Orkunt Sabuncu
#

from atomrule import *
from program import program_atoms
from copy import deepcopy
from random import randint

class CompletedPart:
    
    def __init__(self, e_vars, conj):
	self.existential_vars = e_vars
	self.conj_part = conj
	
    def toTPTPFormat(self):
	evars = ''
	if (len(self.existential_vars) > 0):
	    evars += '? ['
	    for v in self.existential_vars:
		evars += v[1]+','
	    evars = evars[:-1] + ']: '
	    
	conjp = '('
	for c in self.conj_part:
	    conjp += c.toTPTPFormat() + ' & '
	conjp = conjp[:-3] +')'
	
	str = '(' + evars + conjp + ')'
	return str
	    

# deepcopy is used for not changing the original rules while renaming variables
def completeRule(rule):
    eq_a = Atom('==',2)
   
    if (rule.is_fact):
	i = 0
	conj = []
	for a in rule.head.vars:
	    if (a[0] == 'O'):
		l = Literal(eq_a,False,[('V','U_%d'%(i)),a])
		conj[len(conj):] = [l]
	    #elif (a[0] == 'V'):
		# can be problem
		#l = Literal(eq_a,False,[('V','U_%d'%(i)),a])
		#conj[len(conj):] = [l]
	    i += 1
		
	return CompletedPart([],conj)
    
    elif (rule.is_constraint):
	return CompletedPart([],deepcopy(rule.body))
    
    else:
	i = 0
	conj = []
	for a in rule.head.vars:
	    if (a[0] == 'O'):
		l = Literal(eq_a,False,[('V','U_%d'%(i)), a])
		conj[len(conj):] = [l]
            i += 1
	
	conj[len(conj):] = deepcopy(rule.body)
	
	headvars_to_univvars = dict([(v[1],'U_%d'%(i)) for i,v in enumerate(rule.head.vars) if v[0] == 'V'])
	for l in conj:
	    for i,v in enumerate(l.vars):
		if (v[0] == 'V'):
		    l.vars[i] = ('V', headvars_to_univvars.get(v[1],v[1]))
		
	e_vars = list( set(rule.bodyVariables()) - set(rule.headVariables()) )
	
	return CompletedPart(e_vars,conj)
	
	
class CompletedFormula:
    
    def __init__(self):
	self.predicate = None
	self.disj_part = []
	self.universal_vars = []
	
    def setUniversalVars(self):
	self.universal_vars = self.predicate.vars
	
    def renameVars(self):
	vartable = list("ABCDEFGHIJKLMNOPQRSTUVWYZ")
	existvars = [ev[1] for disj in self.disj_part for ev in disj.existential_vars]
	univars_to_something = {}
	
	for v in self.universal_vars:
	    # didn't consider the case we ran out of variable names
	    while True:
	       i = randint(0,len(vartable)-1)
	       if vartable[i] in existvars:
		   continue
	       else:
		   break
	    univars_to_something[v[1]] = vartable[i]
	    vartable[i:i+1] = []
	    
	if (self.predicate):
	    for i,v in enumerate(self.predicate.vars):
		if (v[0] == 'V'):
		    self.predicate.vars[i] = ('V', univars_to_something.get(v[1],v[1]))
	else:
	    for i,v in enumerate(self.universal_vars):
		self.universal_vars[i] = ('V', univars_to_something.get(v[1],v[1]))
			
	for cp in self.disj_part:
	    for l in cp.conj_part:
		for i,v in enumerate(l.vars):
		    if (v[0] == 'V'):
                        l.vars[i] = ('V', univars_to_something.get(v[1],v[1]))
	
    def toTPTPFormat(self, name='cf'):
	str = 'fof('+name+',axiom,\n  '
	if (len(self.universal_vars) > 0):
	    str = str + '! ['
	    for v in self.universal_vars:
		str = str + v[1] + ','
	    str = str[:-1] + ']: '
	
	if (self.predicate):
            str += '( ' + self.predicate.toTPTPFormat() + ' <=> (\n'
	else:
	    str += '( $false <= (\n'

	if (len(self.disj_part) == 0):
	    str += '\t$false ) )).'
	else:
	    for d in self.disj_part:
		str += '\t' + d.toTPTPFormat() + ' |\n'
	    str = str[:-3] + ') )).'
	return str
	    
		
def completeProgram(program,atoms,completed_rules):
    comp_program = []
    for a in atoms.atoms.itervalues():
	if ((a.name == '==') or (a.name == '!=')):
	    continue
	
	comp = CompletedFormula()
	comp.predicate = Literal(a, False, [('V','U_%d'%(i)) for i in range(a.arity)])
	for r in a.inHead:
	    idx = program.index(r)
            comp.disj_part.append(completed_rules[idx][0])
	    completed_rules[idx] = (completed_rules[idx][0],True)
	    
	comp.setUniversalVars()
	comp_program.append(comp)
	
    for i,c in enumerate(completed_rules):
	if (c[1]):
	    continue
	comp = CompletedFormula()
	comp.disj_part.append(c[0])
	comp.universal_vars = list(set(program[i].bodyVariables()))
        
	comp_program.append(comp)
	
    return comp_program


def complete_program(program):
    tmp = map(completeRule,program)
    completed_rules = [(cp,False) for cp in tmp]
    
    return completeProgram(program, program_atoms, completed_rules)
	    
    