#
# Yacc file for the completor
# uses PLY 
#
# Orkunt Sabuncu
#

from lexer import tokens
from node import Node, NodeType
from atomrule import AtomFactory

atomfactory = AtomFactory()


def p_program(p):
    '''program : rules'''
    p[0] = p[1]

def p_empty(p):
    'empty :'
    pass

def p_rules(p):
    '''rules : rules rule '.'
	     | empty'''
    if (len(p) == 4):
	p[2][1] += p[3]
	p[0] = p[1] + [ p[2] ]
    elif (len(p) == 2):
	p[0] = []

def p_rule(p):
    '''rule : head
	    | head IF tail
	    | IF tail'''
    if (len(p) == 2):
	p[0] = [(p[1][0],None), p[1][1]]
	h = p[1][0].getatom()
	a = atomfactory.create(h[0],h[1])
	a.inHead.append(p[0])
    elif (len(p) == 4):
	p[0] = [(p[1][0],p[3][0]), p[1][1]+p[2]+p[3][1]]
	h = p[1][0].getatom()
	a = atomfactory.create(h[0],h[1])
	a.inHead.append(p[0])
    elif (len(p) == 3):
	p[0] = [(None,p[2][0]), p[1]+p[2][1]]
	
def p_head(p):
    '''head : atom'''
    p[0] = p[1]
	    
def p_head_empty(p):
    '''head : empty'''

def p_tail(p):
    '''tail : basic_tail'''
    p[0] = p[1]

def p_basic_tail_1(p):
    '''basic_tail : basic_tail ',' literal
		  | literal'''
    if (len(p) == 4):
	    p[0] = (p[1][0] + [p[3][0]], p[1][1]+p[2]+p[3][1])
    elif (len(p) == 2):
	    p[0] = ([p[1][0]], p[1][1])
	    
def p_basic_tail_2(p):
    '''basic_tail : basic_tail ',' nc_expr
		  | nc_expr'''
    if (len(p) == 4):
	    p[0] = (p[1][0] + [p[3][0]], p[1][1]+p[2]+p[3][1])
    elif (len(p) == 2):
	    p[0] = ([p[1][0]], p[1][1])

def p_literal(p):
    '''literal : basic_literal'''
    p[0] = p[1]

def p_basic_literal(p):
    '''basic_literal : atom
		     | NOT atom'''
    if (len(p) == 2):
	    p[0] = p[1]
    elif (len(p) == 3):
	    p[2][0].naf = True
	    p[0] = (p[2][0], p[1]+' '+p[2][1])
	    
def p_atom(p):
    '''atom : IDENTIFIER '(' arguments ')'
	    | IDENTIFIER'''
    if (len(p) == 5):
	    p[0] = (Node(NodeType.LITERAL,p[1],p[3][0],False), p[1]+p[2]+p[3][1]+p[4])
    elif (len(p) == 2):
	    p[0] = (Node(NodeType.LITERAL,p[1],[],False), p[1])

def p_arguments_arglist(p):
    '''arguments : arglist'''
    p[0] = p[1]
	
def p_arguments_multiargs(p):
    '''arguments : multiargs'''
    p[0] = ([Node(NodeType.MULTARG_EXPR,'',p[1][0],False)], p[1][1])

def p_multiargs(p):
    '''multiargs : arglist 
                 | multiargs ';' arglist'''
    if (len(p) == 2):
	p[0] = ([p[1][0]], p[1][1])
    elif (len(p) == 4):
        p[0] = (p[1][0] + [p[3][0]], p[1][1]+p[2]+p[3][1])    

def p_arglist(p):
    '''arglist : arglist ',' term
	       | term'''
    if (len(p) == 4):
	    p[0] = (p[1][0] + [p[3][0]], p[1][1]+p[2]+p[3][1])
    elif (len(p) == 2):
	    p[0] = ([p[1][0]], p[1][1])
	    
def p_term_range(p):
    '''term : range'''
    p[0] = p[1]
	    
def p_term_atom(p):
    '''term : atom'''
    p[0] = p[1]

def p_term_expr(p):
    '''term : expr'''
    p[0] = p[1]
	
def p_range(p):
    '''range : expr DOTS expr'''
    p[0] = (Node(NodeType.RANGE_EXPR, '', [p[1][0],p[3][0]], False), p[1][1]+p[2]+p[3][1])

def p_expr(p):
    '''expr : '(' expr ')'
	    | constant'''
    if (len(p) == 4):
	p[0] = (p[2][0], p[1]+p[2][1]+p[3])
    elif (len(p) == 2):
	p[0] = p[1]

def p_expr_eq(p):
    '''expr : expr EQ expr'''
    p[0] = (Node(NodeType.LITERAL, '==', [p[1][0],p[3][0]], False), p[1][1]+p[2]+p[3][1])
    
def p_expr_neq(p):
    '''expr : expr NEQ expr'''
    p[0] = (Node(NodeType.LITERAL, '!=', [p[1][0],p[3][0]], False), p[1][1]+p[2]+p[3][1])
	    	    
def p_expr_var(p):
    '''expr : VARIABLE'''
    p[0] = (Node(NodeType.VARIABLE, p[1], [], False),p[1])
    
def p_expr_abs(p):
    '''expr : '|' expr '|' '''
    p[0] = (Node(NodeType.EXPR, '|', [p[2][0]], False), p[1]+p[2][1]+p[3])
    
def p_expr_exp(p):
    '''expr : expr EXP expr '''
    p[0] = (Node(NodeType.EXPR, '**', [p[1][0],p[3][0]], False), p[1][1]+p[2]+p[3][1])
    
def p_expr_plus(p):
    '''expr : expr '+' expr '''
    p[0] = (Node(NodeType.EXPR, '+', [p[1][0],p[3][0]], False), p[1][1]+p[2]+p[3][1])
    
def p_expr_minus(p):
    '''expr : expr '-' expr '''
    p[0] = (Node(NodeType.EXPR, '-', [p[1][0],p[3][0]], False), p[1][1]+p[2]+p[3][1])
    
def p_expr_times(p):
    '''expr : expr '*' expr '''
    p[0] = (Node(NodeType.EXPR, '*', [p[1][0],p[3][0]], False), p[1][1]+p[2]+p[3][1])
        
def p_expr_div(p):
    '''expr : expr '/' expr '''
    p[0] = (Node(NodeType.EXPR, '/', [p[1][0],p[3][0]], False), p[1][1]+p[2]+p[3][1])
    
def p_expr_mod(p):
    '''expr : expr MOD expr '''
    p[0] = (Node(NodeType.EXPR, 'mod', [p[1][0],p[3][0]], False), p[1][1]+p[2]+p[3][1])
    
def p_expr_lt(p):
    '''expr : expr LT expr '''
    p[0] = (Node(NodeType.EXPR, '<', [p[1][0],p[3][0]], False), p[1][1]+p[2]+p[3][1])
    
def p_expr_le(p):
    '''expr : expr LE expr '''
    p[0] = (Node(NodeType.EXPR, '<=', [p[1][0],p[3][0]], False), p[1][1]+p[2]+p[3][1])
    
def p_expr_gt(p):
    '''expr : expr GT expr '''
    p[0] = (Node(NodeType.EXPR, '>', [p[1][0],p[3][0]], False), p[1][1]+p[2]+p[3][1])
    
def p_expr_ge(p):
    '''expr : expr GE expr '''
    p[0] = (Node(NodeType.EXPR, '>=', [p[1][0],p[3][0]], False), p[1][1]+p[2]+p[3][1])
    
def p_expr_assign(p):
    '''expr : expr ASSIGN expr '''
    p[0] = (Node(NodeType.EXPR, 'assign', [p[1][0],p[3][0]], False), p[1][1]+p[2]+p[3][1])



def p_nc_expr(p):
    '''nc_expr : '(' expr ')' '''
    p[0] = (p[2][0], p[1]+p[2][1]+p[3])

def p_nc_expr_eq(p):
    '''nc_expr : expr EQ expr'''
    p[0] = (Node(NodeType.LITERAL, '==', [p[1][0],p[3][0]], False), p[1][1]+p[2]+p[3][1])
    
def p_nc_expr_neq(p):
    '''nc_expr : expr NEQ expr'''
    p[0] = (Node(NodeType.LITERAL, '!=', [p[1][0],p[3][0]], False), p[1][1]+p[2]+p[3][1])

def p_nc_expr_abs(p):
    '''nc_expr : '|' expr '|' '''
    p[0] = (Node(NodeType.EXPR, '|', [p[2][0]], False), p[1]+p[2][1]+p[3])
    
def p_nc_expr_plus(p):
    '''nc_expr : expr '+' expr '''
    p[0] = (Node(NodeType.EXPR, '+', [p[1][0],p[3][0]], False), p[1][1]+p[2]+p[3][1])
    
def p_nc_expr_minus(p):
    '''nc_expr : expr '-' expr '''
    p[0] = (Node(NodeType.EXPR, '-', [p[1][0],p[3][0]], False), p[1][1]+p[2]+p[3][1])
    
def p_nc_expr_times(p):
    '''nc_expr : expr '*' expr '''
    p[0] = (Node(NodeType.EXPR, '*', [p[1][0],p[3][0]], False), p[1][1]+p[2]+p[3][1])
        
def p_nc_expr_div(p):
    '''nc_expr : expr '/' expr '''
    p[0] = (Node(NodeType.EXPR, '/', [p[1][0],p[3][0]], False), p[1][1]+p[2]+p[3][1])
    
def p_nc_expr_mod(p):
    '''nc_expr : expr MOD expr '''
    p[0] = (Node(NodeType.EXPR, 'mod', [p[1][0],p[3][0]], False), p[1][1]+p[2]+p[3][1])
    
def p_nc_expr_lt(p):
    '''nc_expr : expr LT expr '''
    p[0] = (Node(NodeType.EXPR, '<', [p[1][0],p[3][0]], False), p[1][1]+p[2]+p[3][1])
    
def p_nc_expr_le(p):
    '''nc_expr : expr LE expr '''
    p[0] = (Node(NodeType.EXPR, '<=', [p[1][0],p[3][0]], False), p[1][1]+p[2]+p[3][1])
    
def p_nc_expr_gt(p):
    '''nc_expr : expr GT expr '''
    p[0] = (Node(NodeType.EXPR, '>', [p[1][0],p[3][0]], False), p[1][1]+p[2]+p[3][1])
    
def p_nc_expr_ge(p):
    '''nc_expr : expr GE expr '''
    p[0] = (Node(NodeType.EXPR, '>=', [p[1][0],p[3][0]], False), p[1][1]+p[2]+p[3][1])
    
def p_nc_expr_assign(p):
    '''nc_expr : expr ASSIGN expr '''
    p[0] = (Node(NodeType.EXPR, 'assign', [p[1][0],p[3][0]], False), p[1][1]+p[2]+p[3][1])
    

def p_constant_idn(p):
    '''constant : IDENTIFIER'''
    p[0] = (Node(NodeType.CONST_ID, p[1], [], False), p[1])
		
		
def p_constant_num(p):
    '''constant : NUMBER'''
    p[0] = (Node(NodeType.NUMBER, p[1], [], False), p[1])


def p_error(p):
    print "Syntax error in input!"

