#
# Node class
# Used in parsing
#
# Orkunt Sabuncu
#

from atomrule import Atom
import subprocess
from sys import stderr

''' NodeType enum'''
class NodeType:
    LITERAL, NUMBER, CONST_ID, VARIABLE, RANGE_EXPR, MULTARG_EXPR, EXPR = range(7)

    type_lookup = { LITERAL:'LITERAL',
                    NUMBER:'NUMBER',
		    CONST_ID:'CONST_ID',
		    VARIABLE:'VARIABLE',
		    RANGE_EXPR:'RANGE_EXPR',
		    MULTARG_EXPR:'MULTARG_EXPR',
		    EXPR:'EXPR' }
		    
class Node:
    def __init__(self,type,value,args,naf):
	self.type = type
	self.val = value
	self.args = args
	self.naf = naf
	
    def __repr__(self):
	type_str = NodeType.type_lookup[self.type]
	
	return 'Node('+type_str+','+str(self.val)+','+str(map(str,self.args))+','+str(self.naf)+')'
    	    
    def getatom(self):
        if (len(self.args) == 0):
            return (self.val,0)
        elif (self.args[0].type == NodeType.MULTARG_EXPR):
            return (self.val,len(self.args[0].args[0]))
        else:
            return (self.val,len(self.args))

    
''' Expand a node's arguments according to the range and multi-argument expressions
    returns a list of arguments'''    
def expandNode(n, lookup):
    ans =  [[]]
    for arg in n.args:    
	if (arg.type == NodeType.RANGE_EXPR):
	    start = int(eval(arg.args[0].val, None, lookup))
	    end = int(eval(arg.args[1].val, None, lookup))
	    tmp = []
	    for i in range(start,end+1):
                tmp[len(tmp):] = map(lambda x:x+[Node(NodeType.NUMBER,str(i),[],False)], ans)
	    ans = tmp
	    
        elif (arg.type == NodeType.MULTARG_EXPR):
	    tmp = []
	    for i in arg.args:
		tmp[len(tmp):] = expandNode(Node(NodeType.LITERAL,'dummy',i,False), lookup)
            if (ans[0] == []):
		ans = tmp
	    else:
		ans[len(ans):] = tmp
	    
	else :
            ans = map(lambda x: x+[arg], ans)
	    
    return ans
	
''' Post process a parsed program to expand range and multi-argument expressions.'''
def expandRule((head,body), lookup):
    head_expand = [[]]
    body_expand = []
    if (head):
	# not a constraint rule
	head_expand = expandNode(head, lookup)

    if (body):
	# not a fact
	for i in range(len(body)):
	    t = expandNode(body[i], lookup)
	    if (len(t) > 1):
		body_expand[len(body_expand):] = [Node(body[i].type, body[i].val, x, body[i].naf)  for x in t]
	    else:
		body_expand[len(body_expand):] = [body[i]]
    
    # now form the new rules
    if (body_expand == []):
	body_expand = None
    if (not head):
	return [(None,body_expand)]
    return [(Node(head.type, head.val, x, head.naf), body_expand) for x in head_expand]
	

def doesNodeContainExpr(node):
    #print "NODE: ", node
    if (node.type == NodeType.EXPR):
	return True
    if (node.type == NodeType.MULTARG_EXPR):
	for args in node.args:
	    for a in args:
		r = doesNodeContainExpr(a)
		if (r):
		    return True
    else:		
	for a in node.args:
	    r = doesNodeContainExpr(a)
	    if (r):
		return True
    return False
	
def doesRuleContainExpr((head,body)):
    #print "RULE: " , (head,body)
    if (head):
        r = doesNodeContainExpr(head)
        if (r):
            return True
    if (not body):
	return False
    for b in body:
	r = doesNodeContainExpr(b)
	if (r):
	    return True
    return False
	
def addRulesToSubProgram(a, selected_atoms, atoms, subprg):
    if ( (a.name,a.arity) in selected_atoms ):
	return
    
    for [r,str] in a.inHead:
	subprg.append(str)
    
    selected_atoms.add( (a.name,a.arity) )
    
    for [r,str] in a.inHead:
	if (not r[1]):
	    continue
	for b in r[1]:
	    tmp = b.getatom()
	    if ((not tmp in selected_atoms) and (tmp in atoms)):
		newa = atoms[tmp]
		addRulesToSubProgram(newa, selected_atoms, atoms, subprg)
			
''' Find EXPR Nodes and form a subprogram '''
def constructProgramForExprNodes(ptree, atoms):
    selected_atoms = set()
    subprogram = []
    dummy = 0
        
    for r in ptree:
	if (doesRuleContainExpr(r[0])):
            if (r[0][0]):
		tmp = r[0][0].getatom()
		a = atoms[tmp]
		addRulesToSubProgram(a, selected_atoms, atoms, subprogram)
	    else:
		++dummy
		dummyatom = Atom(str(dummy),dummy)
		dummyatom.inHead.append(r)
		addRulesToSubProgram(dummyatom, selected_atoms, atoms, subprogram)
		
    return subprogram
		
''' prepare a file with the sub-program '''
def writeSubProgramToFile(subprogram,filename):
    f = open(filename,'w')
    for r in subprogram:
	f.write('%s\n' % (r) )
    f.close()
    
def callLparse(filename, lookup):
    callstr = 'lparse -t '
    for (k,v) in lookup.items():
	callstr += '-c '+k+'='+str(v)+' '
    callstr += filename
	
    outfile = filename+'.out'
    f = open(outfile,'w')
    ret = subprocess.call(callstr.split(), stdout=f)
    f.close()
    
    if (ret != 0):
	raise Exception(callstr) 
    return outfile
    
def secondPhaseParsing(ptree, atomfactory, filename, lookup):
    subprg = constructProgramForExprNodes(ptree, atomfactory.atoms)
    if (len(subprg) == 0):
	# there is no need for parsing again, just expand the nodes and form the last parsetree
	print >> stderr, 'No need for second parsing'
        return ''
    
    subprg_filename = filename+'.sublp'
    writeSubProgramToFile(subprg, subprg_filename)
    subprg_out = callLparse(subprg_filename, lookup)
    
    # read output
    f = open(subprg_out,'r')
    out= f.readlines()
    # delete the last compute statement
    out[-1:] = []
    f.close()
    
    second_lp_str = ''
    for [r,rstr] in ptree:
	if (rstr not in subprg):
	    second_lp_str += '%s\n' %(rstr)
	    #f.write('%s\n' % (rstr))
	    
	    
    #f.write("%\n% subprogram's output\n%\n")
    second_lp_str += "\n%\n% subprogram's output\n%\n"
    for r in out:
	#f.write(r)
	second_lp_str += r
    
    second_lp = filename+'.2'
    f = open(second_lp,'w')
    f.write(second_lp_str)
    f.close()
    
    return second_lp_str
    
    

