#
# Generate program rules from the parse tree
#
# Orkunt Sabuncu
#

from node import NodeType
from atomrule import AtomFactory, Literal, Rule
from sys import stderr
import options

program_atoms = AtomFactory()

'''
    This is ugly way of handling objects.
    When the term is a fucntion, atom's arguments are used for the function arguments
    Generally there will be objects which are represented by atoms with arity 0.
    I know this is ugly !!!
'''
program_objects = AtomFactory()

def populate_objects_innodes(nodes, insideLiteral, objects):
    for n in nodes:
	if (n.type == NodeType.LITERAL):
	    if (insideLiteral):
		# check whether it is in program_atoms
		if (n.getatom() in program_atoms.atoms):
		    print >> stderr, 'Atom %s is an argument of another atom.'%(n.val)
                else:
		    k = n.getatom()
		    objects.create(k[0],k[1])
		    
	    #recursively call for the arguments
	    for a in n.args:
		populate_objects_innodes([a], True, objects)
	elif (n.type == NodeType.NUMBER or n.type == NodeType.CONST_ID):
	    k = n.getatom()
	    objects.create(k[0],k[1])
	    return
	else:
	    assert n.type == NodeType.VARIABLE
	    return 
	
def populate_objects(ptree):
    program_objects.reset()
    
    for rule in ptree:
	nodes = []
	if (rule[0]):
	    nodes.append(rule[0])
	if (rule[1]):
	    nodes.extend(rule[1])
	    
	populate_objects_innodes(nodes, False, program_objects)
	

def createRule((head,body),atoms):
    r = Rule()
    
    if (not head):
	r.is_constraint = True
    else:
	assert head.type == NodeType.LITERAL
	tmp_vars = []
	for v in head.args:
	    if (v.type == NodeType.VARIABLE):
		tmp_vars[len(tmp_vars):] = [('V',v.val)]
	    elif (v.type == NodeType.NUMBER):
		tmp_vars[len(tmp_vars):] = [('O',v.val)]
	    elif (v.type == NodeType.LITERAL):
		# can be object or literal ???
		# may need evaluation
		tmp_vars[len(tmp_vars):] = [('O',v.val)]
            
	a = atoms.create(head.val, len(head.args))
	l = Literal(a,head.naf,tmp_vars)
	r.head = l
	a.addToinHead(r)
    
    if (not body):
	r.is_fact = True
	r.body = None
    else:
	# now process the body nodes
	for b in body:
	    assert b.type == NodeType.LITERAL
	    tmp_vars = []
	    for v in b.args:
		if (v.type == NodeType.VARIABLE):
                    tmp_vars[len(tmp_vars):] = [('V',v.val)]
		elif (v.type == NodeType.NUMBER):
                    tmp_vars[len(tmp_vars):] = [('O',v.val)]
		elif (v.type == NodeType.LITERAL):
                    # can be object or literal ???
                    # may need evaluation
                    tmp_vars[len(tmp_vars):] = [('O',v.val)]
		
	    a = atoms.create(b.val, len(b.args))
	    l = Literal(a, b.naf, tmp_vars)
	    r.addToBody(l)
	    a.addToinBody(r)
    
    return r

'''
    Numeric objects are not handled by FM-Darwin
    Add a 'n_' prefix to number objects
'''
def change_numeric_objects(prog):
    for r in prog:
	if not r.is_constraint:
	    for i,a in enumerate(r.head.vars):
		if (a[0] == 'O') and a[1].isdigit():
		    r.head.vars[i] = ('O','n_'+a[1])
		    
	if not r.is_fact:
	    for bl in r.body:
                for i,a in enumerate(bl.vars):
                    if (a[0] == 'O') and a[1].isdigit():
                        bl.vars[i] = ('O','n_'+a[1])
			
    # update the program_objects
    for k in program_objects.atoms.keys():
	if k[0].isdigit():
	    del program_objects.atoms[k]
	    program_objects.create('n_'+k[0],0)
	    
def construct_program(ptree):
    # reset the program_atoms
    # maybe it's a better idea not to reset it. In that case more than one parse trees 
    # can be combined to form a single program.
    program_atoms.reset()
    
    p = map(lambda x: createRule(x, program_atoms), ptree)
    
    populate_objects(ptree)
    program_atoms.differentiateAtomsWithSameName()
    
    if options.options.rename_numbers:
        change_numeric_objects(p)
    
    return p

