#
# Prepare the unique name axioms
# Sort inference is handled by this module
#
# Orkunt Sabuncu
#

from atomrule import *
from completion import CompletedFormula, CompletedPart
from options import options

'''
    From David Eppstein's code
    http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/221251
'''
class UnionFind:
    '''Union Find data structure. Modified from Josiah Carlson's code,
    http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/215912
    to allow arbitrarily many arguments in unions, use [] syntax for finds,
    and eliminate unnecessary code.'''

    def __init__(self):
        self.weights = {}
        self.parents = {}

    def __getitem__(self, object):
        '''Find the root of the set that an object is in.
        Object must be hashable; previously unknown objects become new singleton sets.'''

        # check for previously unknown object
        if object not in self.parents:
            self.parents[object] = object
            self.weights[object] = 1
            return object
        
        # find path of objects leading to the root
        path = [object]
        root = self.parents[object]
        while root != path[-1]:
            path.append(root)
            root = self.parents[root]
        
        # compress the path and return
        for ancestor in path:
            self.parents[ancestor] = root
        return root

    def union(self, *objects):
        '''Find the sets containing the given objects and merge them all.'''
        roots = [self[x] for x in objects]
        heaviest = max([(self.weights[r],r) for r in roots])[1]
        for r in roots:
            if r != heaviest:
                self.weights[heaviest] += self.weights[r]
                self.parents[r] = heaviest


def infer_sorts(prog, patoms, pobjects):
    uf = UnionFind()
    l = []
    
    for a in patoms.atoms.keys():
	if (a[0] == '!=' or a[0] == '=='):
	    continue
	l.extend([ (a[0],i) for i in range(1,a[1]+1) ])
    
    for a in pobjects.atoms.keys():
	l.append(a)
    
    for i in l:
	uf[i]
	
    if (options.unsorted):
	for i in range(len(l)-1):
	    uf.union(l[i], l[i+1])
	
	sorts = {}
	for i in uf.parents.keys():
	    if i in pobjects.atoms:
		if (uf[i] in sorts):
		    sorts[uf[i]].append(i[0])
		else:
		    sorts[uf[i]] = [i[0]]
        
        return (sorts, uf.parents)
    
    for r in prog:
	lits = []
	vartable = {}
	if (r.is_fact):
	    lits.append(r.head)
	elif (r.is_constraint):
	    lits.extend(r.body)
	else:
	    lits.append(r.head)
	    lits.extend(r.body)
	    
	# put the equality literal at the end of list so that variables are bounded 
	# when we process equality literal
	for l in lits:
	    if (l.atom.name == '=='):
		lits.remove(l)
		lits.append(l)
	
	#print 'RULE START'
	
	for l in [l for l in lits if l.atom.name != '!=']:
	    #print l.atom.name, l.vars
	    
	    if (l.atom.name == '=='):
		left, right = None, None
		if (l.vars[0][0] == 'V'):
		    left = vartable[l.vars[0][1]]
		elif (l.var[0][0] == 'O'):
		    left = (l.vars[0][1],0)
		if (l.vars[1][0] == 'V'):
		    right = vartable[l.vars[1][1]]
		elif (l.var[1][0] == 'O'):
		    right = (l.vars[1][1],0)
		    
		uf.union(left,right)
		continue
		        
	    for (i,a) in enumerate(l.vars):
		if (a[0] == 'O'):
		    uf.union( (l.atom.name,i+1), (a[1],0) )
		elif (a[0] == 'V'):
		    if (a[1] not in vartable):
			vartable[a[1]] = (l.atom.name,i+1)
		    else:
			uf.union( (l.atom.name,i+1), vartable[a[1]] )
	
	#print 'RULE END'
	
    sorts = {}
    for i in uf.parents.keys():
	if i in pobjects.atoms:
	    if (uf[i] in sorts):
		sorts[uf[i]].append(i[0])
	    else:
		sorts[uf[i]] = [i[0]]
        
    #return sorts.values()
    return (sorts, uf.parents)
	
	
def generate_naiveuna_for_sort(sort):
    eq_a = Atom('==',2)
    axioms = []
    
    for i in range(0,len(sort)):
        for j in range(i+1,len(sort)):
	    l = Literal(eq_a, False, [('O',sort[i]), ('O',sort[j])])
	    cp = CompletedPart([], [l])
	    f = CompletedFormula()
	    f.disj_part.append(cp)
	    axioms.append(f)
    
    return axioms
	
def generate_nlognuna_for_sort(sort):
    i = 1
    uatoms = []
    axioms = []
    
    while True:
	uatoms.append(Atom('unq%s_%d'%(sort[0],i),1))
	if len(sort) > (1 << i):
	    i += 1
	else:
	    break
	
    for i,e in enumerate(sort):
	for j,ua in enumerate(uatoms):
	    if ((1<<j) & i) > 0 :
		l = Literal(ua, True, [('O',e)])
		cp = CompletedPart([], [l])
		f = CompletedFormula()
		f.disj_part.append(cp)
		axioms.append(f)
	    else:
		l = Literal(ua, False, [('O',e)])
		cp = CompletedPart([], [l])
		f = CompletedFormula()
		f.disj_part.append(cp)
		axioms.append(f)
	
    return axioms
	    
def generate_una(sorts):
    all = []
    
    for s in sorts:
	if (options.una == 'Naive'):
            all.extend(generate_naiveuna_for_sort(s))
	elif (options.una == 'Nlogn'):
	    all.extend(generate_nlognuna_for_sort(s))
	
    return all
	

